/***********************************************************************
tcping.exe -- A tcp probe utility
Copyright (C) 2005 Eli Fulkerson

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

----------------------------------------------------------------------

Other license terms may be negotiable.  Contact the author if you would
like a copy that is licensed differently.

Contact information (as well as this program) lives at http://www.elifulkerson.com

----------------------------------------------------------------------

This application includes public domain code from the Winsock Programmer's FAQ:
  http://www.tangentsoft.net/wskfaq/
... and a big thank you to the maintainers and contributers therein.

***********************************************************************/

#include "ws-util.h"

#include <winsock2.h>

#include <iostream>
#include <string.h>
#include <time.h>

//#include <sys/timeb.h> // for ftime

#define NUM_PROBES 4

using namespace std;

////////////////////////////////////////////////////////////////////////
// Prototypes

u_long LookupAddress(const char* pcHost);
SOCKET EstablishConnection(u_long nRemoteAddr, u_short nPort, int ping_interval);
LARGE_INTEGER cpu_frequency;
LARGE_INTEGER response_timer1;
LARGE_INTEGER response_timer2;

int CTRL_C_ABORT;

void controlc() {
	cerr << "Control-C" << endl;
	CTRL_C_ABORT = 1;
	//return true;
}

int DoWinsock( const char* pcHost, int nPort, int times_to_ping, int ping_interval, int include_timestamp, int beep_mode, int ping_timeout )
{
	SetConsoleCtrlHandler((PHANDLER_ROUTINE)&controlc, TRUE);
	CTRL_C_ABORT = 0;
	/*
	int start_time = 0;
	int end_time = 0;
	int start_milltime = 0;
	int end_milltime = 0;
	int total_ms = 0;			// total number of milliseconds per pass
	*/
	
	int number_of_pings = 0;	// total number of tcpings issued
	double running_total_ms = 0;	// running total of values of pings... divide by number_of_pings for avg
	double lowest_ping = 50000;		// lowest ping in series ... starts high so it will drop
	double max_ping = 0;			// highest ping in series
	
	int success_counter = 0;
	int failure_counter = 0;


	//struct _timeb timebuffer;

	int loopcounter = 0;			// number of probes to send
	
	// Timestamp variablees
	time_t rawtime;
	struct tm * timeinfo;
	char dateStr [11];
	char timeStr [9];

	int beep_flag = -1;  // 0 for we're down, 1 for we're up
	double response_time;

	cout << endl;

    // Find the server's address
    u_long nRemoteAddress = LookupAddress(pcHost);
    if (nRemoteAddress == INADDR_NONE) {
		cerr << endl << "Could not find host " << pcHost << ", aborting." << endl;
        return 3;
    }
    in_addr Address;
    memcpy(&Address, &nRemoteAddress, sizeof(u_long)); 
    
	while (loopcounter < times_to_ping || (times_to_ping == -1 && CTRL_C_ABORT == 0) )
	{
		
		if (include_timestamp == 1) {
			//_strdate( dateStr);
			_strtime( timeStr );
			time ( &rawtime );
			timeinfo = localtime ( &rawtime );
			strftime(dateStr, 11, "%Y:%m:%d",timeinfo);
			cout << dateStr << " " << timeStr << " ";
		}


		cout << "Probing " << inet_ntoa(Address) << ":" << nPort << "/tcp - ";

		// apparently... QueryPerformanceCounter isn't thread safe unless we do this
		SetThreadAffinityMask(GetCurrentThread(),1); 
		
		// start the timer right before we do the connection
		QueryPerformanceFrequency((LARGE_INTEGER *)&cpu_frequency);
		QueryPerformanceCounter((LARGE_INTEGER *) &response_timer1);

	    // Connect to the server
	    SOCKET sd = EstablishConnection(nRemoteAddress, htons(nPort), ping_timeout);

		// grab the timeout as early as possible
		QueryPerformanceCounter((LARGE_INTEGER *) &response_timer2);
		

	    if (sd == INVALID_SOCKET) {
			cerr <<  WSAGetLastErrorMessage("");
			failure_counter++;
		    	
		        if (beep_mode == 4 || beep_mode == 1 || (beep_mode == 3 && beep_flag == 1)) {
				cout << char(7) << char(7);
				beep_flag = 0;
			}
	    }
		else {
			cout << "Port is open";
			success_counter++;
			
			if (beep_mode == 4 || beep_mode == 2 || (beep_mode == 3 && beep_flag == 0)) {
				cout << char(7);
				beep_flag = 1;
			}
  	}
		    // Shut connection down
	if (ShutdownConnection(sd)) {
		// room here for connection shutdown success check...
	}
		else {
		// room here for connection shutdown failure check...
	}
		

        response_time=( (double) ( (response_timer2.QuadPart - response_timer1.QuadPart) * (double) 1000.0 / (double) cpu_frequency.QuadPart) );
		
		cout.precision(3);
		cout.setf(ios::showpoint);
		cout.setf(ios::fixed);
		
		cout << " - time=" << response_time << "ms" << endl;
		
		// Calculate the statistics...
		number_of_pings++;

		if (sd != INVALID_SOCKET) {
			running_total_ms += response_time;

			if (response_time < lowest_ping)
			{
				lowest_ping = response_time;
			}

			if (response_time > max_ping)
			{
				max_ping = response_time;
			}
		}

		loopcounter++;
		if (loopcounter == times_to_ping) {
			break;
		}
		
		int zzz = 0;
		int wakeup = (ping_interval * 1000) - response_time;
		if (wakeup > 0 )
		{
			while (zzz < wakeup && CTRL_C_ABORT ==0) {
				Sleep(10);
				zzz += 10;
			}
		}
	}
	
	
	cout << endl << "Ping statistics for " << inet_ntoa(Address) << ":" << nPort << endl;
	cout << "     " << number_of_pings << " probes sent. "<< endl;
	cout << "     " << success_counter << " successful, " << failure_counter << " failed." << endl;
	if (success_counter > 0) {
		if ( failure_counter > 0) {
			cout << "Approximate trip times in milli-seconds (successful connections only):" << endl;
		} else {
			cout << "Approximate trip times in milli-seconds:" << endl;
		}
		cout.precision(3);
		cout << "     Minimum = " << lowest_ping << "ms, Maximum = " << max_ping << "ms, Average = " << running_total_ms/number_of_pings << "ms" <<endl;
	} else {
		cout << "Was unable to connect, cannot provide trip statistics." << endl;
	}
		
	// report our total, abject failure.
	if (success_counter == 0) {
		return 1;
	}
	
	// return our intermittent failure
	if (success_counter > 0 && failure_counter > 0) {
		return 2;	
	}
		
    return 0;
}


//// LookupAddress /////////////////////////////////////////////////////
// Given an address string, determine if it's a dotted-quad IP address
// or a domain address.  If the latter, ask DNS to resolve it.  In
// either case, return resolved IP address.  If we fail, we return
// INADDR_NONE.

u_long LookupAddress(const char* pcHost)
{
    u_long nRemoteAddr = inet_addr(pcHost);
    if (nRemoteAddr == INADDR_NONE) {
        // pcHost isn't a dotted IP, so resolve it through DNS
        hostent* pHE = gethostbyname(pcHost);
        if (pHE == 0) {
            return INADDR_NONE;
        }
        nRemoteAddr = *((u_long*)pHE->h_addr_list[0]);
    }

    return nRemoteAddr;
}


//// EstablishConnection ///////////////////////////////////////////////
// Connects to a given address, on a given port, both of which must be
// in network byte order.  Returns newly-connected socket if we succeed,
// or INVALID_SOCKET if we fail.

SOCKET EstablishConnection(u_long nRemoteAddr, u_short nPort, int ping_timeout)
{
	
	LARGE_INTEGER timer1;
	LARGE_INTEGER timer2;
	LARGE_INTEGER cpu_freq;
	
	double time_so_far;

	// Create a stream socket
	SOCKET sd = socket(AF_INET, SOCK_STREAM, 0);

	//-------------------------
	// Set the socket I/O mode: In this case FIONBIO
	// enables or disables the blocking mode for the 
	// socket based on the numerical value of iMode.
	// If iMode = 0, blocking is enabled; 
	// If iMode != 0, non-blocking mode is enabled.
	u_long iMode = 1;
	ioctlsocket(sd, FIONBIO, &iMode);
		
        sockaddr_in sinRemote;
        sinRemote.sin_family = AF_INET;
        sinRemote.sin_addr.s_addr = nRemoteAddr;
        sinRemote.sin_port = nPort;

        QueryPerformanceFrequency((LARGE_INTEGER *)&cpu_freq);
        QueryPerformanceCounter((LARGE_INTEGER *) &timer1);
	
	connect(sd, (sockaddr*)&sinRemote, sizeof(sockaddr_in));
    
	char *sendy = ".";
	int sendstatus = 1000;
	
	bool done = false;
	
	//for (int i=0;i<ping_timeout;i+=10) {
	while (!done) {
		sendstatus = send(sd, sendy, 0, 0);
		//cout << sendstatus;

		// one error code is if you send a send of size 0, the other is if you actually send data.  Leaving it here for now,
		// but have left it in a "send nothing" mode rather than sending the "."
		if (sendstatus == 1 || sendstatus == 0) {
			return sd;
		}
		
		QueryPerformanceCounter((LARGE_INTEGER *) &timer2);
		
		time_so_far = ( (double) ( (timer2.QuadPart - timer1.QuadPart) * (double) 1000.0 / (double) cpu_freq.QuadPart) );
		
		if (time_so_far >= ping_timeout) {
			done = true;
		} else {		
			Sleep(1);
		}
	}

	return INVALID_SOCKET;
    //return sd;
}